@@ -7,9 +7,6 @@ from aenet_gpr.inout.io_print import *
 
 
 def main():
-    device = "cpu"
-    if torch.cuda.is_available():
-        device = "cuda:0"
 
     io_print_header()
 
@@ -20,7 +17,9 @@ def main():
         input_file = "train.in"
     input_param = read_train_in(input_file)
     input_param.update_soap_param()
-    input_param.device = device
+
+    if input_param.device == 'gpu' and torch.cuda.is_available():
+        input_param.device = "cuda:0"
 
     # if input_param.verbose:
     #     io_input_reading(input_param)
