@@ -1,5 +1,5 @@
 from .gpr_iterative import GaussianProcess
-from .gpr_batch import GaussianProcess
+from .gpr_batch import GaussianProcess, numerical_descriptor_gradient
 # from .gpr_batch_internal import GaussianProcessInternal
 from .pytorch_kernel import FPKernel, FPKernelNoforces
 # from .pytorch_kernel_internal import FPKernelInternal
@@ -8,4 +8,5 @@ from .calculator import GPRCalculator
 from .prior import ConstantPrior
 
 __all__ = ["gpr_iterative", "gpr_batch", "gpr_batch_internal", "pytorch_kernel", "pytorch_kernel_internal", "pytorch_kerneltypes", "calculator", "prior",
-           "GaussianProcess", "FPKernel", "FPKernelNoforces", "SquaredExp", "GPRCalculator", "ConstantPrior"]
+           "GaussianProcess", "numerical_descriptor_gradient", "FPKernel", "FPKernelNoforces", "SquaredExp", "GPRCalculator", "ConstantPrior",
+           ]
