@@ -1,5 +1,7 @@
 import torch
 import numpy as np
+
+from copy import deepcopy
 from joblib import Parallel, delayed
 
 from aenet_gpr.src.prior import ConstantPrior
@@ -7,86 +9,164 @@ from aenet_gpr.src.pytorch_kernel import FPKernel
 from aenet_gpr.util.prepare_data import get_N_batch, get_batch_indexes_N_batch
 
 
-def _central_diff_task(atoms, i, j, delta, model, num_layers):
+# def _central_diff_task(atoms, i, j, delta, model, num_layers):
+#     """
+#     Desciptor gradient for each atom i, and cartexian coordinate j(x/y/z)
+#     """
+#     atoms_f = deepcopy(atoms)
+#     atoms_f.positions[i, j] += delta
+#     desc_f = model.get_descriptors(atoms_f, num_layers=num_layers)
+#
+#     atoms_b = deepcopy(atoms)
+#     atoms_b.positions[i, j] -= delta
+#     desc_b = model.get_descriptors(atoms_b, num_layers=num_layers)
+#
+#     grad_ij = (desc_f - desc_b) / (2 * delta)
+#     return i, j, grad_ij
+#
+#
+# def numerical_descriptor_gradient_parallel(atoms, model, delta=1e-4, num_layers=-1, n_jobs=1, dtype=torch.float32):
+#     """
+#     Args:
+#         atoms (ase.Atoms):
+#         model: MACE calculator
+#         delta (float): displacement (Ang)
+#         num_layers (int): MACE interaction layers number
+#         n_jobs (int): pararell job number
+#         dtype (torch.dtype):
+#
+#     Returns:
+#         desc (torch.Tensor): (n_atoms, descriptor_dim)
+#         grad (torch.Tensor): (n_atoms, n_atoms, 3, descriptor_dim)
+#     """
+#
+#     desc = model.get_descriptors(atoms, num_layers=num_layers)
+#     n_atoms, D = desc.shape
+#
+#     # Job list
+#     jobs = [(atoms, i, j, delta, model, num_layers) for i in range(n_atoms) for j in range(3)]
+#
+#     # Pararell execution
+#     results = Parallel(n_jobs=n_jobs, prefer="threads")(delayed(_central_diff_task)(*job) for job in jobs)
+#     results.sort(key=lambda x: (x[0], x[1]))
+#
+#     grad = torch.empty((n_atoms, n_atoms, 3, D), dtype=dtype)
+#     for i, j, grad_ij in results:
+#         grad[:, i, j, :] = torch.tensor(grad_ij, dtype=dtype)
+#
+#     return torch.tensor(desc, dtype=dtype), grad
+
+
+def numerical_descriptor_gradient(atoms, model, delta=1e-4, num_layers=-1, n_jobs=-1, dtype='float32'):
     """
-    Desciptor gradient for each atom i, and cartexian coordinate j(x/y/z)
-    """
-    atoms_f = atoms.copy()
-    atoms_f.positions[i, j] += delta
-    desc_f = model.get_descriptors(atoms_f, num_layers=num_layers)
+    Optimized batch version - pre-generates all perturbed positions and processes them efficiently
 
-    atoms_b = atoms.copy()
-    atoms_b.positions[i, j] -= delta
-    desc_b = model.get_descriptors(atoms_b, num_layers=num_layers)
-
-    grad_ij = (desc_f - desc_b) / (2 * delta)
-    return i, j, grad_ij
-
-
-def numerical_descriptor_gradient_parallel(atoms, model, delta=1e-4, num_layers=-1, n_jobs=-1, dtype=torch.float32):
-    """
     Args:
         atoms (ase.Atoms):
         model: MACE calculator
         delta (float): displacement (Ang)
         num_layers (int): MACE interaction layers number
-        n_jobs (int): pararell job number
-        dtype (torch.dtype):
+        n_jobs (int): number of processes
+        dtype (dtype):
 
     Returns:
         desc (torch.Tensor): (n_atoms, descriptor_dim)
         grad (torch.Tensor): (n_atoms, n_atoms, 3, descriptor_dim)
     """
-
+    # Get original descriptor
     desc = model.get_descriptors(atoms, num_layers=num_layers)
     n_atoms, D = desc.shape
 
-    # Job list
-    jobs = [(atoms, i, j, delta, model, num_layers) for i in range(n_atoms) for j in range(3)]
+    # Pre-allocate gradient array
+    grad = np.empty((n_atoms, n_atoms, 3, D), dtype=dtype)
 
-    # Pararell execution
-    results = Parallel(n_jobs=n_jobs, backend="loky")(delayed(_central_diff_task)(*job) for job in jobs)
-
-    grad = torch.empty((n_atoms, n_atoms, 3, D), dtype=dtype)
-    for i, j, grad_ij in results:
-        grad[:, i, j, :] = torch.tensor(grad_ij, dtype=dtype)
-
-    return torch.tensor(desc, dtype=dtype), grad
-
-
-def numerical_descriptor_gradient(atoms, model, delta=1e-4, num_layers=-1):
-    """
-    atoms: ASE Atoms object
-    model: MACE calculator (mace_mp)
-    delta: displacement for numerical differentiation
-
-    Returns:
-        gradient: shape (N_atoms, 3, descriptor_dim)
-    """
-    atoms = atoms.copy()
-    n_atoms = len(atoms)
-
-    desc_0 = model.get_descriptors(atoms, num_layers=num_layers)  # shape: (n_atoms, D)
-    D = desc_0.shape[1]
-
-    grad = torch.empty((n_atoms, n_atoms, 3, D))
+    # Get original positions once
+    original_positions = atoms.get_positions()
 
+    # Pre-create all perturbations
+    perturbations = []
     for i in range(n_atoms):
         for j in range(3):  # x, y, z
-            # forward step
-            atoms_f = atoms.copy()
-            atoms_f.positions[i, j] += delta
-            desc_f = model.get_descriptors(atoms_f, num_layers=num_layers)
+            # Forward perturbation
+            pos_f = original_positions.copy()
+            pos_f[i, j] += delta
+            perturbations.append(('f', i, j, pos_f))
 
-            # backward step
-            atoms_b = atoms.copy()
-            atoms_b.positions[i, j] -= delta
-            desc_b = model.get_descriptors(atoms_b, num_layers=num_layers)
+            # Backward perturbation
+            pos_b = original_positions.copy()
+            pos_b[i, j] -= delta
+            perturbations.append(('b', i, j, pos_b))
 
-            # central difference
-            grad[:, i, j, :] = (desc_f - desc_b) / (2 * delta)
+    if n_jobs == 1:
+        # Compute all descriptors
+        all_descriptors = []
+        for direction, i, j, positions in perturbations:
+            atoms_temp = deepcopy(atoms)
+            atoms_temp.set_positions(positions)
+            desc_temp = model.get_descriptors(atoms_temp, num_layers=num_layers)
+            all_descriptors.append((direction, i, j, desc_temp))
+    else:
+        # --- Parallelize descriptor computation ---
+        def compute_descriptor(direction, i, j, positions):
+            atoms_temp = deepcopy(atoms)
+            atoms_temp.set_positions(positions)
+            desc_temp = model.get_descriptors(atoms_temp, num_layers=num_layers)
+            return direction, i, j, desc_temp
 
-    return desc_0, grad
+        all_descriptors = Parallel(n_jobs=n_jobs, prefer="threads")(
+            delayed(compute_descriptor)(direction, i, j, positions)
+            for (direction, i, j, positions) in perturbations
+        )
+
+    # Reorganize and compute gradients
+    desc_dict = {}
+    for direction, i, j, desc_temp in all_descriptors:
+        key = (i, j)
+        if key not in desc_dict:
+            desc_dict[key] = {}
+        desc_dict[key][direction] = desc_temp
+
+    # Compute central differences
+    for (i, j), descs in desc_dict.items():
+        grad[:, i, j, :] = (descs['f'] - descs['b']) / (2 * delta)
+
+    return desc, grad
+
+
+# def numerical_descriptor_gradient(atoms, model, delta=1e-4, num_layers=-1, dtype='float32'):
+#     """
+#     Args:
+#         atoms (ase.Atoms):
+#         model: MACE calculator
+#         delta (float): displacement (Ang)
+#         num_layers (int): MACE interaction layers number
+#         dtype (torch.dtype):
+#
+#     Returns:
+#         desc (torch.Tensor): (n_atoms, descriptor_dim)
+#         grad (torch.Tensor): (n_atoms, n_atoms, 3, descriptor_dim)
+#     """
+#
+#     desc = model.get_descriptors(atoms, num_layers=num_layers)
+#     n_atoms, D = desc.shape
+#
+#     grad = np.empty((n_atoms, n_atoms, 3, D), dtype=dtype)
+#     for i in range(n_atoms):
+#         for j in range(3):  # x, y, z
+#             # forward step
+#             atoms_f = deepcopy(atoms)
+#             atoms_f.positions[i, j] += delta
+#             desc_f = model.get_descriptors(atoms_f, num_layers=num_layers)
+#
+#             # backward step
+#             atoms_b = deepcopy(atoms)
+#             atoms_b.positions[i, j] -= delta
+#             desc_b = model.get_descriptors(atoms_b, num_layers=num_layers)
+#
+#             # central difference
+#             grad[:, i, j, :] = (desc_f - desc_b) / (2 * delta)
+#
+#     return desc, grad
 
 
 def apply_force_mask(F, atoms_mask):
@@ -119,6 +199,7 @@ class GaussianProcess(object):
 
     kernel: Defaults to the Squared Exponential kernel with derivatives
     '''
+
     def __init__(self, hp=None, prior=None, prior_update=True, kerneltype='sqexp',
                  scale=0.4, weight=1.0, noise=1e-6, noisefactor=0.5,
                  use_forces=True, images=None, function=None, derivative=None,
@@ -136,6 +217,7 @@ class GaussianProcess(object):
             self.torch_data_type = torch.float64
 
         self.device = device
+
         self.soap_param = soap_param
         self.mace_param = mace_param
         self.descriptor = descriptor
@@ -178,33 +260,34 @@ class GaussianProcess(object):
         elif self.descriptor == 'mace':
             if self.mace_param.get('system') == "materials":
                 try:
-                    from mace_descriptor.calculators import mace_mp
+                    from mace.calculators import mace_mp
                 except ImportError:
                     raise ImportError(
-                        "The 'joblib', 'mace' and 'mace-descriptor' packages are required for using pre-trained MACE descriptors.\n"
+                        "The 'joblib' and 'mace' packages are required for using pre-trained MACE descriptors.\n"
                         "Please install it by running:\n\n"
                         "    pip install joblib\n"
                         "    pip install mace-torch\n"
-                        "    pip install mace-descriptor\n\n"
-                        "Note: This is a lightweight fork of the original MACE (mace-torch) package, "
-                        "designed exclusively for descriptor extraction."
                     )
 
+                # Check and set device
+                if torch.cuda.is_available():
+                    mace_device = "cuda:0"
+                else:
+                    mace_device = "cpu"
+                    print("[Warning] CUDA device not available. MACE descriptor computation may be slow on CPU.")
+
                 self.mace = mace_mp(model=self.mace_param.get('model'),
-                                    device=self.device)
+                                    device=mace_device)
 
             else:
                 try:
-                    from mace_descriptor.calculators import mace_off
+                    from mace.calculators import mace_off
                 except ImportError:
                     raise ImportError(
-                        "The 'joblib', 'mace' and 'mace-descriptor' packages are required for using pre-trained MACE descriptors.\n"
+                        "The 'joblib' and 'mace' packages are required for using pre-trained MACE descriptors.\n"
                         "Please install it by running:\n\n"
                         "    pip install joblib\n"
                         "    pip install mace-torch\n"
-                        "    pip install mace-descriptor\n\n"
-                        "Note: This is a lightweight fork of the original MACE (mace-torch) package, "
-                        "designed exclusively for descriptor extraction."
                     )
 
                 self.mace = mace_off(model=self.mace_param.get('model'),
@@ -286,26 +369,27 @@ class GaussianProcess(object):
                                                return_descriptor=True,
                                                n_jobs=self.soap_param.get('n_jobs'))
 
-            dfp_dr = torch.as_tensor(dfp_dr, dtype=self.torch_data_type).to(self.device)  # (Ndata, Ncenters, Natom, 3, Natom*3)
+            dfp_dr = torch.as_tensor(dfp_dr, dtype=self.torch_data_type).to(
+                self.device)  # (Ndata, Ncenters, Natom, 3, Natom*3)
             fp = torch.as_tensor(fp, dtype=self.torch_data_type).to(self.device)  # (Ndata, Ncenters, Natom*3)
 
         elif self.descriptor == 'mace':
+
             fp = []
             dfp_dr = []
             for image in images:
-                # fp__, dfp_dr__ = self.mace.get_descriptors_with_jacobian(image)
-                # fp.append(fp__)
-                # dfp_dr.append(dfp_dr__)
-
-                if self.device == 'cpu':
-                    fp__, dfp_dr__ = numerical_descriptor_gradient_parallel(image, self.mace)
-                else:
-                    fp__, dfp_dr__ = numerical_descriptor_gradient(image, self.mace)
-                fp.append(fp__)
-                dfp_dr.append(dfp_dr__)
+                fp__, dfp_dr__ = numerical_descriptor_gradient(image,
+                                                               self.mace,
+                                                               delta=self.mace_param.get("delta"),
+                                                               num_layers=self.mace_param.get("num_layers"),
+                                                               n_jobs=self.mace_param.get("mace_n_jobs"),
+                                                               dtype=self.data_type)
+                fp.append(torch.tensor(fp__, dtype=self.torch_data_type, device=self.device))
+                dfp_dr.append(torch.tensor(dfp_dr__, dtype=self.torch_data_type, device=self.device))
 
             fp = torch.stack(fp).to(dtype=self.torch_data_type, device=self.device)  # (Ndata, Natom, Ndescriptor)
-            dfp_dr = torch.stack(dfp_dr).to(dtype=self.torch_data_type, device=self.device)  # (Ndata, Natom, Natom, 3, Ndescriptor)
+            dfp_dr = torch.stack(dfp_dr).to(dtype=self.torch_data_type,
+                                            device=self.device)  # (Ndata, Natom, Natom, 3, Ndescriptor)
 
         else:
             fp = []
@@ -339,12 +423,14 @@ class GaussianProcess(object):
             fp = torch.as_tensor(fp, dtype=self.torch_data_type).to(self.device)  # (Ncenters, Natom*3)
 
         elif self.descriptor == 'mace':
-            if self.device == 'cpu':
-                fp, dfp_dr = numerical_descriptor_gradient_parallel(image, self.mace)
-            else:
-                fp, dfp_dr = numerical_descriptor_gradient(image, self.mace)
-            fp = fp.to(dtype=self.torch_data_type, device=self.device)  # (Natom, Ndescriptor)
-            dfp_dr = dfp_dr.to(dtype=self.torch_data_type, device=self.device)  # (Natom, Natom, 3, Ndescriptor)
+            fp, dfp_dr = numerical_descriptor_gradient(image,
+                                                       self.mace,
+                                                       delta=self.mace_param.get("delta"),
+                                                       num_layers=self.mace_param.get("num_layers"),
+                                                       n_jobs=self.mace_param.get("mace_n_jobs"),
+                                                       dtype=self.data_type)
+            fp = torch.tensor(fp, dtype=self.torch_data_type, device=self.device)  # (Natom, Ndescriptor)
+            dfp_dr = torch.tensor(dfp_dr, dtype=self.torch_data_type, device=self.device)  # (Natom, Natom, 3, Ndescriptor)
 
         else:
             fp = torch.as_tensor(image.get_positions(wrap=False).reshape(-1), dtype=self.torch_data_type).to(
@@ -423,8 +509,9 @@ class GaussianProcess(object):
 
                 pred, kernel = self.eval_data_batch(eval_fp, eval_dfp_dr)
                 E_hat[eval_x_indexes[i][0]:eval_x_indexes[i][1]] = pred[0:data_per_batch]
-                F_hat[eval_x_indexes[i][0]:eval_x_indexes[i][1], :] = apply_force_mask(F=pred[data_per_batch:].view(data_per_batch, -1),
-                                                                                       atoms_mask=self.atoms_mask)
+                F_hat[eval_x_indexes[i][0]:eval_x_indexes[i][1], :] = apply_force_mask(
+                    F=pred[data_per_batch:].view(data_per_batch, -1),
+                    atoms_mask=self.atoms_mask)
 
             return E_hat, F_hat.view((Ntest, self.Natom, 3)), None, None
 
@@ -438,8 +525,9 @@ class GaussianProcess(object):
 
                 pred, kernel = self.eval_data_batch(eval_fp, eval_dfp_dr)
                 E_hat[eval_x_indexes[i][0]:eval_x_indexes[i][1]] = pred[0:data_per_batch]
-                F_hat[eval_x_indexes[i][0]:eval_x_indexes[i][1], :] = apply_force_mask(F=pred[data_per_batch:].view(data_per_batch, -1),
-                                                                                       atoms_mask=self.atoms_mask)
+                F_hat[eval_x_indexes[i][0]:eval_x_indexes[i][1], :] = apply_force_mask(
+                    F=pred[data_per_batch:].view(data_per_batch, -1),
+                    atoms_mask=self.atoms_mask)
 
                 var = self.eval_variance_batch(get_variance=get_variance,
                                                eval_fp=eval_fp,
@@ -448,8 +536,9 @@ class GaussianProcess(object):
                 std = torch.sqrt(torch.diagonal(var))
 
                 unc_e[eval_x_indexes[i][0]:eval_x_indexes[i][1]] = std[0:data_per_batch] / self.weight
-                unc_f[eval_x_indexes[i][0]:eval_x_indexes[i][1], :] = apply_force_mask(F=std[data_per_batch:].view(data_per_batch, -1),
-                                                                                       atoms_mask=self.atoms_mask)
+                unc_f[eval_x_indexes[i][0]:eval_x_indexes[i][1], :] = apply_force_mask(
+                    F=std[data_per_batch:].view(data_per_batch, -1),
+                    atoms_mask=self.atoms_mask)
 
             return E_hat, F_hat.view((Ntest, self.Natom, 3)), unc_e, unc_f.view((Ntest, self.Natom, 3))
 
