@@ -74,7 +74,10 @@ class ReferenceData(object):
 
         if mace_param is None:
             self.mace_param = {'system': "materials",
-                               'model': "small"}
+                               'model': "small",
+                               'delta': 1e-4,
+                               'num_layers': -1,
+                               'mace_n_jobs': 1}
         else:
             self.mace_param = mace_param
 
